@@ -1,4 +1,4 @@
-#  Jnterface.pm
+#  Interface.pm
 #    - providing an object-oriented approach to interacting with GnuPG
 #
 #  Copyright (C) 2000 Frank J. Tobin <ftobin@cpan.org>
@@ -12,7 +12,8 @@
 #
 
 package GnuPG::Interface;
-use Any::Moose;
+use Moo;
+use MooX::late;
 with qw(GnuPG::HashInit);
 
 use English qw( -no_match_vars );
@@ -26,14 +27,41 @@ use IO::Handle;
 use Math::BigInt try => 'GMP';
 use GnuPG::Options;
 use GnuPG::Handles;
+use Scalar::Util 'tainted';
 
-$VERSION = '0.44';
+$VERSION = '1.02';
 
-has $_ => (
+has passphrase => (
     isa     => 'Any',
     is      => 'rw',
-    clearer => 'clear_' . $_,
-) for qw(call passphrase);
+    clearer => 'clear_passphrase',
+);
+
+has call => (
+    isa     => 'Any',
+    is      => 'rw',
+    trigger => 1,
+    clearer => 'clear_call',
+);
+
+# NB: GnuPG versions
+#
+# There are now two supported versions of GnuPG: legacy 1.4 and stable 2.2
+# They are detected and each behave slightly differently.
+#
+# When using features specific to branches, check that the system's
+# version of gpg corresponds to the branch.
+#
+# legacy: 1.4
+# stable: >= 2.2
+#
+# You can find examples of version comparison in the tests.
+has version => (
+    isa      => 'Str',
+    is       => 'ro',
+    reader   => 'version',
+    writer   => '_set_version',
+);
 
 has options => (
     isa        => 'GnuPG::Options',
@@ -48,9 +76,7 @@ sub gnupg_call { shift->call(@_); }
 
 sub BUILD {
     my ( $self, $args ) = @_;
-
-    $self->hash_init( call => 'gpg' );
-    $self->hash_init(%$args);
+    $self->hash_init( call => 'gpg', %$args );
 }
 
 struct(
@@ -61,6 +87,12 @@ struct(
     }
 );
 
+# Update version if "call" is updated
+sub _trigger_call {
+    my ( $self, $gpg ) = @_;
+    $self->_set_version( $self->_version() );
+}
+
 #################################################################
 # real worker functions
 
@@ -76,17 +108,18 @@ sub wrap_call( $% ) {
     $handles->stdout('>&STDOUT') unless $handles->stdout();
     $handles->stderr('>&STDERR') unless $handles->stderr();
 
-    # so call me sexist; English just doen't cope well
-    my $needs_passphrase_handled_for_him
-        = ( $self->passphrase() and not $handles->passphrase() ) ? 1 : 0;
+    $self->passphrase("\n") unless $self->passphrase();
 
-    if ($needs_passphrase_handled_for_him) {
+    my $needs_passphrase_handled
+        = ( $self->passphrase() =~ m/\S/ and not $handles->passphrase() ) ? 1 : 0;
+
+    if ($needs_passphrase_handled) {
         $handles->passphrase( IO::Handle->new() );
     }
 
     my $pid = $self->fork_attach_exec(%args);
 
-    if ($needs_passphrase_handled_for_him) {
+    if ($needs_passphrase_handled) {
         my $passphrase_handle = $handles->passphrase();
         print $passphrase_handle $self->passphrase();
         close $passphrase_handle;
@@ -105,6 +138,16 @@ sub fork_attach_exec( $% ) {
     my ( $self, %args ) = @_;
 
     my $handles = $args{handles} or croak 'no GnuPG::Handles passed';
+    my $use_loopback_pinentry = 0;
+
+    # Don't use loopback pintentry for legacy (1.4) GPG
+    #
+    # Check that $version is populated before running cmp_version. If
+    # we are invoked as part of BUILD to populate $version, then any
+    # methods that depend on $version will fail. We don't care about
+    # loopback when we're called just to check gpg version.
+    $use_loopback_pinentry = 1
+      if ($handles->passphrase() && $self->version && $self->cmp_version($self->version, '2.2') > 0 );
 
     # deprecation support
     $args{commands} ||= $args{gnupg_commands};
@@ -120,6 +163,8 @@ sub fork_attach_exec( $% ) {
         = ref $args{command_args}
         ? @{ $args{command_args} }
         : ( $args{command_args} || () );
+    unshift @command_args, "--"
+        if @command_args and $command_args[0] ne "--";
 
     my %fhs;
     foreach my $fh_name (
@@ -290,11 +335,23 @@ sub fork_attach_exec( $% ) {
             $self->options->$option($fileno);
         }
 
+        my @args = $self->options->get_args();
+
+        # Get around a bug in 2.2, see also https://dev.gnupg.org/T4667
+        # this covers both --delete-secret-key(s) and --delete-secret-and-public-key(s)
+        if ( $self->version && $self->cmp_version( $self->version, 2.2 ) >= 0 && $commands[0] =~ /^--delete-secret-.*keys?$/ ) {
+            push @args, '--yes';
+        }
+
+        push @args, '--pinentry-mode', 'loopback'
+          if $use_loopback_pinentry;
+
         my @command = (
-            $self->call(), $self->options->get_args(),
+            $self->call(), @args,
             @commands,     @command_args
         );
 
+        local $ENV{PATH} if tainted $ENV{PATH};
         exec @command or die "exec() error: $ERRNO";
     }
 
@@ -415,7 +472,7 @@ sub get_keys {
     while (<$stdout>) {
         my $line = $_;
         chomp $line;
-        my @fields = split ':', $line;
+        my @fields = split ':', $line, -1;
         next unless @fields > 3;
 
         my $record_type = $fields[0];
@@ -592,7 +649,7 @@ sub get_keys {
           my ($pos, $size, $data) = @fields[ 1,2,3 ];
           $current_key->pubkey_data->[$pos+0] = Math::BigInt->from_hex('0x'.$data);
         }
-        elsif ( $record_type ne 'tru' ) {
+        elsif ( $record_type ne 'tru' and $record_type ne 'grp' ) {
             warn "unknown record type $record_type";
         }
     }
@@ -656,10 +713,17 @@ sub encrypt( $% ) {
 
 sub encrypt_symmetrically( $% ) {
     my ( $self, %args ) = @_;
-    return $self->wrap_call(
+    # Strip the homedir and put it back after encrypting;
+    my $homedir = $self->options->homedir;
+    $self->options->clear_homedir
+        unless $self->cmp_version($self->version, '2.2') >= 0;
+    my $pid = $self->wrap_call(
         %args,
         commands => ['--symmetric']
     );
+    $self->options->homedir($homedir)
+        unless $self->cmp_version($self->version, '2.2') >= 0;
+    return $pid;
 }
 
 sub sign( $% ) {
@@ -745,6 +809,46 @@ sub send_keys( $% ) {
     );
 }
 
+sub search_keys( $% ) {
+    my ( $self, %args ) = @_;
+    return $self->wrap_call(
+        %args,
+        commands => ['--search-keys']
+    );
+}
+
+sub _version {
+    my ( $self ) = @_;
+
+    my $out = IO::Handle->new;
+    my $handles = GnuPG::Handles->new( stdout => $out );
+    my $pid = $self->wrap_call( commands => [ '--no-options', '--version' ], handles => $handles );
+    my $line = $out->getline;
+    $line =~ /(\d+\.\d+\.\d+)/;
+
+    my $version = $1;
+    unless ($self->cmp_version($version, '2.2') >= 0 or
+        ($self->cmp_version($version, '1.4') >= 0 and $self->cmp_version($version, '1.5') < 0 )) {
+        croak "GnuPG Version 1.4 or 2.2+ required";
+    }
+    waitpid $pid, 0;
+
+    return $version;
+}
+
+sub cmp_version($$) {
+    my ( $self, $a, $b ) = (@_);
+    my @a = split '\.', $a;
+    my @b = split '\.', $b;
+    @a > @b
+        ? push @b, (0) x (@a-@b)
+        : push @a, (0) x (@b-@a);
+    for ( my $i = 0; $i < @a; $i++ ) {
+        return $a[$i] <=> $b[$i] if $a[$i] <=> $b[$i];
+    }
+    return 0;
+}
+
 sub test_default_key_passphrase() {
     my ($self) = @_;
 
@@ -780,7 +884,7 @@ sub test_default_key_passphrase() {
 
     # all we realy want to check is the status fh
     while (<$status>) {
-        if (/^\[GNUPG:\]\s*GOOD_PASSPHRASE/) {
+        if (/^\[GNUPG:\]\s*(GOOD_PASSPHRASE|SIG_CREATED)/) {
             waitpid $pid, 0;
             return 1;
         }
@@ -805,8 +909,8 @@ GnuPG::Interface - Perl interface to GnuPG
   # A simple example
   use IO::Handle;
   use GnuPG::Interface;
-  
-  # settting up the situation
+
+  # setting up the situation
   my $gnupg = GnuPG::Interface->new();
   $gnupg->options->hash_init( armor   => 1,
 			      homedir => '/home/foobar' );
@@ -824,7 +928,7 @@ GnuPG::Interface - Perl interface to GnuPG
   # Now we'll go about encrypting with the options already set
   my @plaintext = ( 'foobar' );
   my $pid = $gnupg->encrypt( handles => $handles );
-  
+
   # Now we write to the input of GnuPG
   print $input @plaintext;
   close $input;
@@ -846,7 +950,7 @@ decryption, verification, and key-listing parsing.
 =head2 How Data Member Accessor Methods are Created
 
 Each module in the GnuPG::Interface bundle relies
-on Any::Moose to generate the get/set methods
+on Moo to generate the get/set methods
 used to set the object's data members.
 I<This is very important to realize.>  This means that
 any data member which is a list has special
@@ -878,6 +982,19 @@ B<clearsign>.  One then interacts with with the handles
 appropriately, as described in
 L<perlipc/"Bidirectional Communication with Another Process">.
 
+=head1 GnuPG Versions
+
+As of this version of GnuPG::Interface, there are two supported
+versions of GnuPG: 1.4.x and 2.2.x. The
+L<GnuPG download page|https://gnupg.org/download/index.html> has
+updated information on the currently supported versions.
+
+GnuPG released 2.0 and 2.1 versions in the past and some packaging
+systems may still provide these if you install the default C<gpg>,
+C<gnupg>, C<gnupg2>, etc. packages. This modules supports only
+version 2.2.x, so you may need to find additional package
+repositories or build from source to get the updated version.
+
 =head1 OBJECT METHODS
 
 =head2 Initialization Methods
@@ -928,6 +1045,8 @@ initialization of data members.
 
 =item send_keys( % )
 
+=item search_keys( % )
+
 These methods each correspond directly to or are very similar
 to a GnuPG command described in L<gpg>.  Each of these methods
 takes a hash, which currently must contain a key of B<handles>
@@ -975,13 +1094,25 @@ and standard error will be tied to the running program's standard error,
 standard output, or standard error.  If the B<status> or B<logger> handle
 is not defined, this channel of communication is never established with GnuPG,
 and so this information is not generated and does not come into play.
+
 If the B<passphrase> data member handle of the B<handles> object
 is not defined, but the the B<passphrase> data member handle of GnuPG::Interface
 object is, GnuPG::Interface will handle passing this information into GnuPG
-for the user as a convience.  Note that this will result in
+for the user as a convenience.  Note that this will result in
 GnuPG::Interface storing the passphrase in memory, instead of having
 it simply 'pass-through' to GnuPG via a handle.
 
+If neither the B<passphrase> data member of the GnuPG::Interface nor
+the B<passphrase> data member of the B<handles> object is defined,
+then GnuPG::Interface assumes that access and control over the secret
+key will be handled by the running gpg-agent process.  This represents
+the simplest mode of operation with the GnuPG "stable" suite (version
+2.2 and later).  It is also the preferred mode for tools intended to
+be user-facing, since the user will be prompted directly by gpg-agent
+for use of the secret key material.  Note that for programmatic use,
+this mode requires the gpg-agent and pinentry to already be correctly
+configured.
+
 =back
 
 =head2 Other Methods
@@ -1009,6 +1140,10 @@ while signing a short message using the values of
 the B<passphrase> data member, and the default
 key specified in the B<options> data member.
 
+=item version()
+
+Returns the version of GnuPG that GnuPG::Interface is running.
+
 =back
 
 
@@ -1096,10 +1231,15 @@ The following setup can be done before any of the following examples:
 
   $gnupg->options->hash_init( armor    => 1,
                               recipients => [ 'ftobin@uiuc.edu',
-                                              '0xABCD1234' ],
-                              meta_interactive( 0 ),
+                                              '0xABCD1234ABCD1234ABCD1234ABCD1234ABCD1234' ],
+                              meta_interactive => 0 ,
                             );
 
+   $gnupg->options->debug_level(4);
+
+   $gnupg->options->logger_file("/tmp/gnupg-$$-decrypt-".time().".log");
+
+
 =head2 Encrypting
 
   # We'll let the standard error of GnuPG pass through
@@ -1110,7 +1250,7 @@ The following setup can be done before any of the following examples:
 
   my $handles = GnuPG::Handles->new( stdin    => $input,
                                      stdout   => $output );
-   
+
   # this sets up the communication
   # Note that the recipients were specified earlier
   # in the 'options' data member of the $gnupg object.
@@ -1141,7 +1281,7 @@ The following setup can be done before any of the following examples:
 				   );
 
   # indicate our pasphrase through the
-  # convience method
+  # convenience method
   $gnupg->passphrase( $passphrase );
 
   # this sets up the communication
@@ -1186,7 +1326,7 @@ The following setup can be done before any of the following examples:
   # a file written to disk
   # Make sure you "use IO::File" if you use this module!
   my $cipher_file = IO::File->new( 'encrypted.gpg' );
-   
+
   # this sets up the communication
   my $pid = $gnupg->decrypt( handles => $handles );
 
@@ -1202,9 +1342,9 @@ The following setup can be done before any of the following examples:
   close $input;
   close $cipher_file;
 
-  my @plaintext    = <$output>;   # reading the output
-  my @error_output = <$error>;    # reading the error
-  my @status_info  = <$status_fh> # read the status info
+  my @plaintext    = <$output>;    # reading the output
+  my @error_output = <$error>;     # reading the error
+  my @status_info  = <$status_fh>; # read the status info
 
   # clean up...
   close $output;
@@ -1218,20 +1358,20 @@ The following setup can be done before any of the following examples:
   # This time we'll just let GnuPG print to our own output
   # and read from our input, because no input is needed!
   my $handles = GnuPG::Handles->new();
-  
-  my @ids = ( 'ftobin', '0xABCD1234' );
+
+  my @ids = ( 'ftobin', '0xABCD1234ABCD1234ABCD1234ABCD1234ABCD1234' );
 
   # this time we need to specify something for
   # command_args because --list-public-keys takes
   # search ids as arguments
   my $pid = $gnupg->list_public_keys( handles      => $handles,
                                       command_args => [ @ids ] );
-  
+
    waitpid $pid, 0;
 
 =head2 Creating GnuPG::PublicKey Objects
 
-  my @ids = [ 'ftobin', '0xABCD1234' ];
+  my @ids = [ 'ftobin', '0xABCD1234ABCD1234ABCD1234ABCD1234ABCD1234' ];
 
   my @keys = $gnupg->get_public_keys( @ids );
 
@@ -1246,7 +1386,7 @@ The following setup can be done before any of the following examples:
       command_args => [ qw( test/key.1.asc ) ],
       handles      => $handles,
     );
-    
+
     my @out = <$handles->stdout()>;
     waitpid $pid, 0;
 
@@ -1300,12 +1440,23 @@ one should all B<wait> to clean up GnuPG from the process table.
 
 =head1 BUGS
 
+=head2 Large Amounts of Data
+
 Currently there are problems when transmitting large quantities
 of information over handles; I'm guessing this is due
 to buffering issues.  This bug does not seem specific to this package;
 IPC::Open3 also appears affected.
 
-I don't know yet how well this modules handles parsing OpenPGP v3 keys.
+=head2 OpenPGP v3 Keys
+
+I don't know yet how well this module handles parsing OpenPGP v3 keys.
+
+=head2 RHEL 7 Test Failures
+
+Testing with the updates for version 1.00 we saw intermittent test failures
+on RHEL 7 with GnuPG version 2.2.20. In some cases the tests would all pass
+for several runs, then one would fail. We're unable to reliably reproduce
+this so we would be interested in feedback from other users.
 
 =head1 SEE ALSO
 
@@ -1316,16 +1467,18 @@ L<GnuPG::SecretKey>,
 L<gpg>,
 L<perlipc/"Bidirectional Communication with Another Process">
 
+=head1 LICENSE
+
+This module is free software; you can redistribute it and/or modify it
+under the same terms as Perl itself.
+
 =head1 AUTHOR
 
-GnuPg::Interface is currently maintained by Jesse Vincent <jesse@cpan.org>.  
+GnuPG::Interface is currently maintained by Best Practical Solutions <BPS@cpan.org>.
 
 Frank J. Tobin, ftobin@cpan.org was the original author of the package.
 
 =cut
 
-
-__PACKAGE__->meta->make_immutable;
-
 1;
 
